### Ethernet switching hub (switch) simulator (UDP w/o select) ###
#
# Public domain software
# (When using Python 2, part of this program must be commented/uncommented)
#
# Coded by Yasusi Kanada
# 2013-5-19, 24, 6-10; 2014-5-25, 6-8; 2015-5-11, 6-21; 2016-6-25

import optparse, socket, time
from ether import SMAC, DMAC, printPacket


BUF_SIZE = 3100		# > 2 * 1500

host = socket.gethostbyname(socket.gethostname())


### MAC address table handler ###

MACtable = {}

def snapMACtable(MACaddr, Message):
    entry = MACtable[MACaddr]
    print("{} {} {} {} {} {}".\
          format(Message, " MAC:", MACaddr, " port:", entry["port"],
                 " time:", entry["time"]))
    return

def dumpMACtable():
    print("\n\n\n\n\n\n\n\n\n\n\n\n")
    print("{} {} {}".format("\nMACaddr     ", "Port ", "Time"))
    count = 0
    for MACaddr in MACtable.keys():
        count = count + 1
        entry = MACtable[MACaddr]
        print("{} {} {}".format(MACaddr, entry["port"], entry["time"]))
    if count == 0:
        print("Empty")

def setMACtable(MACaddr, sourcePort, now):
    oldEntry = MACtable.get(MACaddr)
    MACtable[MACaddr] = { "port": sourcePort, "time": now }
    if opts.monitorLevel >= 1 and oldEntry == None:
        snapMACtable(MACaddr, "Newly learned:")
    return MACtable[MACaddr]

def getMACtable(MACaddr):
    return MACtable.get(MACaddr)

def timedOut(tableEntry, now):
    time = now - tableEntry["time"] - opts.timeout > 0
    if opts.monitorLevel >= 2:
        print("{} {} table:{} timedout:{} {}".\
              format("Timed out" if time > 0 else "Not timed out",
                     now, tableEntry["time"], opts.timeout, time))
    return time > 0

def cleanMACtable(now):		# Delete (only) one dead entry
    for MACaddr in MACtable.keys():
        if timedOut(MACtable[MACaddr], now):
            MACtable.pop(MACaddr)
            return


### Packet handler ###

def sendPacket(packet, destPort):
    sock, addr = vport2remote[destPort]
    return sock.sendto(packet, addr) // 2

def printForwarding(message, dmac):
    print("{} {} {} {} {}".format("@@@", message, "- DMAC:", dmac, "@@@"))

def printSending(sendBytes, destPort):
    print("{} {} {}".format(sendBytes, "bytes sent through port", destPort))


### Packet forwarding ###

def forward(packet, sourcePort, ports):
    smac = SMAC(packet)
    now = time.time()
    if opts.monitorLevel >= 1:
        printPacket(packet, now, "Received:")

    # Learn MAC
    sourceEntry = setMACtable(smac, sourcePort, now)

    # Forward packet
    dmac = DMAC(packet)
    destEntry = getMACtable(dmac)
    if destEntry == None:				# Flood packet:
       if opts.monitorLevel >= 1:
           printForwarding("Flood packet", dmac)
       for destPort in ports:
           if sourcePort != destPort:
               sendBytes = sendPacket(packet, destPort)
               if opts.monitorLevel >= 2:
                   printSending(sendBytes, destPort)
    else:						# Switch packet:
        if opts.monitorLevel >= 1:
            printForwarding("Switch packet", dmac)
        destPort = destEntry["port"]
        sendBytes = sendPacket(packet, destPort)
        if opts.monitorLevel >= 2:
            printSending(sendBytes, destPort)


### Main program ###

def main():
    global opts, vport2remote, host

    # Parse command line
    parser = optparse.OptionParser()
    for i in range(0, 8):
        str_i = str(i);
        parser.add_option("--lp" + str_i, "--lport" + str_i,
                          dest = "lport" + str_i, type = "int",
                          help = ("Local (my) network port id for link " + str_i))
        parser.add_option("--rp" + str_i, "--rport" + str_i,
                          dest = "rport" + str_i, type = "int",
                          help = ("Remote (next) network port id for link " + str_i))
    parser.add_option("--nports", dest = "nports", type = "int",
                      help = ("Number of local network ports"))
    parser.add_option("--timeout", dest = "timeout", type = "int",
                      help = "Learning timeout time")
    parser.add_option("--monitorLevel", "--monitor",
                      dest = "monitorLevel", type="int",
                      help = ("Monotoring level (0: Silent, 1: Sending/Receiving 2: Debug)"))
    parser.add_option("--dumpMAC", dest = "dumpMAC", action = "store_true")
    parser.set_defaults(lport0 = 4000, lport1 = 4001, lport2 = 4002, lport3 = 4003,
                        lport4 = 4004, lport5 = 4005, lport6 = 4006, lport7 = 4007,
                        rport0 = 5000, rport1 = 5001, rport2 = 5002, rport3 = 5003,
                        rport4 = 5004, rport5 = 5005, rport6 = 5006, rport7 = 5007,
                        nports = 8, timeout = 300, monitorLevel = 0)
    opts, args = parser.parse_args()

    # Create port lists
    lport = [opts.lport0, opts.lport1, opts.lport2, opts.lport3,
             opts.lport4, opts.lport5, opts.lport6, opts.lport7]

    rport = [opts.rport0, opts.rport1, opts.rport2, opts.rport3,
             opts.rport4, opts.rport5, opts.rport6, opts.rport7]

    # Open sockets
    sock = list([None] * opts.nports)
    sockPort = {}
    for i in range(0, opts.nports):
        sock[i] = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        sock[i].setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sock[i].setblocking(0)
        sock[i].bind((host, lport[i]))
        sockPort[sock[i]] = lport[i]

    # Create link info (virtual port to remote address)
    vport = list([None] * opts.nports)
    vport2remote = {}
    for i in range(0, opts.nports):
        vport[i] = lport[i]
        vport2remote[vport[i]] = (sock[i], (host, rport[i]))

    # Repeat forwarding a packet
    interval = 1
    lastTime = time.time()
    while True:
        for i in range(0, opts.nports):
            sourceSock = sock[i]
            try:
                packet = sourceSock.recv(BUF_SIZE)
            except socket.error:
                pass
            else:
                forward(packet, sockPort[sourceSock], vport)

        now = time.time()
        if now > lastTime + interval:
            cleanMACtable(now)
            if opts.dumpMAC:
                dumpMACtable()
            lastTime = now
        else:
            time.sleep(0.001001)

main()
